
/*
 * sp_update_customer is a procedure that can be called in the upload_update
 * MobiLink event.
 *
 * Standard conflict detection and resolution using MobiLink requires the 
 * following events have been created (for table Customer):
 *      upload_update
 *      upload_fetch
 *      upload_old_row_insert
 *      upload_new_row_insert
 *      resolve_conflict
 *
 * upload_fetch - retrieves the current copy of the row from the consolidated 
 *                database.  MobiLink uses these values to determine if a
 *                conflict has occurred.
 * upload_old_row_insert - If a conflict has occurred MobiLink uses this 
 *                         event to insert the old values uploaded from
 *                         the remote into a temporary table so that the
 *                         resolve_conflict stored procedure has access
 *                         to these values to base business logic rules on
 *                         how to resolve the conflict.
 * upload_new_row_insert - If a conflict has occurred MobiLink uses this 
 *                         event to insert the new values uploaded from
 *                         the remote into a temporary table so that the
 *                         resolve_conflict stored procedure has access
 *                         to these values to base business logic rules on
 *                         how to resolve the conflict.
 * resolve_conflict - A stored procedure that will use:
 *                        a) Values in the "current" consolidated table
 *                        b) Values in the "old" temporary table
 *                        c) Values in the "new" temporary table
 *                    Based on those values, updates the table in the 
 *                    consolidiated database.
 *
 * This is an example of an alternative method to the above.
 * Advantages:
 *    - All the code/business logic is located in 1 stored proceudre.
 *    - No temporary tables are required.
 *
 * Disadvantages
 *    - The consolidated may not provide the ability to determine if an 
 *      update changed any rows.
 *    - This may not work for all supported databases.  For example Sybase
 *      Adaptive Enterprise does not support referencing columns with 
 *      datatypes of LONG / text in the WHERE clause.
 *      
 *      
 * The upload_update event normally accepts the same number of question marks
 * that the table has columns.  So for the Customer table which has 5 columns 
 * (cust_1st_pk, cust_2nd_pk, first_name, last_name, nullable_col, 
 * last_modified), there would be 6 question marks.
 * 
 * call ml_add_table_script( 'v1', 'Customer', 'upload_update', '
 *      UPDATE customer
 *         SET first_name    = ?,
 *             last_name     = ?,
 *             nullable_col  = ?,
 *             last_modified = ?
 *       WHERE cust_1st_pk   = ?
 *         AND cust_2nd_pk   = ?
 *  ' );
 *
 * The upload_update event can also accept additional question marks for 
 * non-primary key columns.  So in this case, it can accept 3 additional
 * question marks, since 2 (of 5) columns in the Customer table are part
 * of the primary key.  See the description in the sp_update_customer.
 *
 * For this example, we will take advantage of this, and call a stored
 * procedure to:
 *      a) Perform the update to the consolidated database.
 *      b) Detect if a conflict has occurred.
 *      c) Resolve the conflict.
 *
 * This only requires 1 event.
 *
 * call ml_add_table_script( 'v1', 'Customer', 'upload_update', '
 *      CALL sp_update_customer( ?, ?, ?, ?, ?, ?, ?, ?, ? )
 *  ' );
 *
 */
CREATE PROCEDURE sp_update_customer(
     IN @new_first_name    varchar(30)
    ,IN @new_last_name     varchar(30)
    ,IN @new_nullable_col  varchar(30)
    ,IN @new_last_modified timestamp
    ,IN @cust_1st_pk       integer
    ,IN @cust_2nd_pk       integer
    ,IN @old_first_name    varchar(30)
    ,IN @old_last_name     varchar(30)
    ,IN @old_nullable_col  varchar(30)
    ,IN @old_last_modified timestamp
)
BEGIN
    DECLARE @current_first_name    varchar(30);
    DECLARE @current_last_name     varchar(30);
    DECLARE @current_nullable_col  varchar(30);
    DECLARE @current_last_modified timestamp;

    -- @new      - these are the values the remote updated the row to
    -- @old      - these are the values that were originally downloaded
    --             to the remote from MobiLink.
    -- @current  - These are the values currently in the consolidated database
    --             before the update from the remote has been applied.
    --
    -- Update the table in the consolidated database.
    -- Instead of only including the primary key columns in the WHERE clause
    -- also include all the OLD columns.  
    -- By providing these additional columns in the WHERE clause we can check
    -- how many rows were updated by the UPDATE statement.
    -- If 1 row was updated:
    --          Then no conflict has occurred and the update was applied
    --          successfully.  We can move on to the next change from the 
    --          remote.
    -- If 0 rows were updated:
    --          Then this row was updated since it was downloaded to the remote
    --          database.  This means it is a CONFLICT.
    --          We can choose to RESOLVE this conflict now.
    --
    -- NOTE:
    -- The use of the COALESCE function below is necessary for all
    -- columns that allow NULL.
    --     WHERE nullable_col = NULL will always return FALSE.
    -- For NULL values, you must use the IS NULL test:
    --     WHERE nullable_col IS NULL will returns TRUE if it is NULL.
    -- The COALESCE function allows us to write a statement that
    -- will work in all cases:
    --     COALESCE(nullable_col, '') = COALESCE(@old_nullable_col, '')
    -- If nullable_col IS NULL, then an empty string is returned.
    -- A test of '' = '' will work and return the results we expect.
    --
    -- NOTE:
    -- If any of the columns are LONG types (ie BLOBs, images, word docs, ...)
    -- then you may have to exclude these columns from the WHERE clause.
    -- This depends on whether your relational database supports the use 
    -- of those columns in a WHERE clause.
    --
    UPDATE customer
       SET first_name                 = @new_first_name,
           last_name                  = @new_last_name,
           nullable_col               = @new_nullable_col,
           last_modified              = @new_last_modified
     WHERE cust_1st_pk                = @cust_1st_pk
       AND cust_2nd_pk                = @cust_2nd_pk
       AND first_name                 = @old_first_name
       AND last_name                  = @old_last_name
       AND COALESCE(nullable_col, '') = COALESCE(@old_nullable_col, '')
       AND last_modified              = @old_last_modified;

    -- The syntax used here depends on your consolidated database.
    -- Sybase Adaptive Server Anywhere/Enterprise and Microsoft SQL Server
    -- supply a global variable, @@rowcount, which indicates how many rows 
    -- were affected by the last operation.  This can be used to determine if 
    -- a conflict has occurred.
    --
    -- If you are using Oracle or DB2, a similar mechanism must exist to use
    -- this method.  Alternatively, you could issue a SELECT statement 
    -- with the above WHERE clause to see if any rows are returned.

    IF( @@rowcount = 0 ) THEN
        -- Useful for debugging, display a message to the minimized
        -- engine window indicating this procedure has fired.
        MESSAGE string( CURRENT TIMESTAMP ) + ' sp_update_customer: '
            + 'Conflict detected for Customer: '
            + string( @cust_1st_pk);

        -- A conflict has been detected.
        -- To resolve, we must:
        --      a) Retrieve the current row values from the consolidated
        --      b) Use business logic to detemine which values to use
        --      c) Update the consolidated database with the final values
        SELECT first_name, last_name, nullable_col, last_modified
          INTO @current_first_name
             , @current_last_name
             , @current_nullable_col
             , @current_last_modified
          FROM customer
         WHERE cust_1st_pk = @cust_1st_pk
           AND cust_2nd_pk = @cust_2nd_pk;

        IF( SQLCODE = 0 ) THEN
            -- Use simple business logic here
            IF( @new_last_modified > @current_last_modified ) THEN
                -- The remote was the most recent update to this row.
                -- So lets use the values it uploaded.
                UPDATE customer
                   SET first_name    = @new_first_name,
                       last_name     = @new_last_name,
                       nullable_col  = @new_nullable_col,
                       last_modified = @new_last_modified
                 WHERE cust_1st_pk   = @cust_1st_pk
                   AND cust_2nd_pk   = @cust_2nd_pk;

                -- Display a message in the minimized engine window
                -- to indicate that the remote has won the conflict.
                MESSAGE string( CURRENT TIMESTAMP ) + ' sp_update_customer: '
                    + 'Remote has won the conflict for Customer: '
                    + string( @cust_1st_pk);
            ELSE
                -- Display a message in the minimized engine window
                -- to indicate that the remote has won the conflict.
                MESSAGE string( CURRENT TIMESTAMP ) + ' sp_update_customer: '
                    + 'Consolidated has won the conflict for Customer: '
                    + string( @cust_1st_pk);
            END IF;

            -- More complex logic could be used and alternatively, 
            -- use some columns from each database.  I could choose to
            -- only update the first_name field and leave the other fields
            -- alone.
        END IF;
    ELSE

        -- Useful for debugging, display a message to the minimized
        -- engine window indicating this procedure has fired.
        MESSAGE string( CURRENT TIMESTAMP ) + ' sp_update_customer: '
            + 'NO conflict detected for Customer: '
            + string( @cust_1st_pk);
    END IF;

END;

-- db_ext:dsnname=cons,type=ASA
